# coffeelint: disable=max_line_length

# require dependencies
debug = require("debug")("omg-producermanager")
_ = require 'lodash'
chalk = require 'chalk'

sockets = require "../../common/node_zmqsockets/"
configloader = require '../../common/node_configloader'
helpers = require './helpers'

#create a new ZapHandler to only allow connections with correct curvekey
new sockets.ZapHandler()
zmqconfig = configloader.zmq
{standardAddress, standardTransport, reqrepbroker, pushpullbroker, pubsubbroker} = zmqconfig

# bind a zmq socket for requests from clients
transport = reqrepbroker?.transport or standardTransport
address = reqrepbroker?.address or standardAddress
port = reqrepbroker?.rep
repPort = "#{transport}://#{address}:#{port}"
replySocket = new sockets.SecureBoundSocket "reqrepbroker", "router", repPort

# bind a zmq socket for pushes from clients
transport = pushpullbroker?.transport or standardTransport
address = pushpullbroker?.address or standardAddress
port = pushpullbroker?.pull
pullAddress = "#{transport}://#{address}:#{port}"
pullSocket = new sockets.SecureBoundSocket "pushpullbroker", "pull", pullAddress

# bind a zmq socket on which new data is published
transport = pubsubbroker.transport or standardTransport
address = pubsubbroker.address or standardAddress
pubAddress = "#{transport}://#{address}:#{pubsubbroker.pub}"
publishSocket = new sockets.SecureBoundSocket "pubsubbroker", "pub", pubAddress

# ### incoming requests are handled here

handleRequest = (producerName, command, producerArgs, callback) ->
  # set flag thats used to indicate if a request to some producer was successfull
  receivedReply = false

  # return error if no producerName was given
  producer = omg[producerName] or false
  unless producer and producer[command]
    if callback
      callback "Producer or command not available"
    return

  # arguments for for producer commands are (should be) stringified JSON.
  # Parse arguments
  if producerArgs
    try
      producerArgs = JSON.parse(producerArgs)
    # make arguments be an empty array if parsing fails ...
    catch error
      producerArgs = []
  # ... or no arguments are given
  else
    producerArgs = []

  # **Callback that is send as last argument to the producer**
  replyCallback = (err, data) ->
    debug "received reply from #{producerName}"
    receivedReply = true
    data = helpers.objectify(data) or ""
    data = (if data then JSON.stringify(data) else "")
    err = err or ""
    status = producer.status or ""
    if callback
      callback err, status, data
  producerArgs.push replyCallback

  # finaly forward altered request to producer
  debug "forwarding request to #{producerName}"
  producer[command].apply producer[command], producerArgs

  # **timeout function is called after 10 seconds**
  requestTimedOut = ->
    # send timeout error if request has not received any data, yet
    unless receivedReply
      if callback
        debug "#{producerName}:#{command} did not respond within 10 seconds"
        callback "timed out"
  setTimeout requestTimedOut, 10 * 1000

# parse messages coming in on reply socket and forward them to request handler
replySocket.on "message", ->
  zmqMessageParts = Array.apply(null, arguments)
  [messageId, nullByte, producerName, command, producerArgs] = zmqMessageParts
  handleRequest producerName, command, producerArgs, (err, status, data) ->
    replySocket.send [messageId, "", producerName, err, status, data]

# parse messages coming in on pull socket and forward them to request handler
pullSocket.on "message", ->
  zmqMessageParts = Array.apply(null, arguments)
  [producerName, command, producerArgs] = zmqMessageParts
  handleRequest producerName, command, producerArgs

# ## Producers

# empty object to which all 'required' producers will be attached
omg = {}

# **All producers that can/must *not* be instantiated, but only "spawned" once**
selfRunningProducers = [
  # [cpuload](../producers/cpuload/lib/omg-cpuload.html)
  'cpuload'
  # [statics](../producers/statics/lib/omg-statics.html)
  'statics'
  # [diskspace](../producers/diskspace/lib/omg-diskspace.html)
  'diskspace'
  # [coredumpcheck](../producers/coredumpcheck/lib/omg-coredumpcheck.html)
  'coredumpcheck'
  # [gatelogs](../producers/gatelogs/lib/omg-gatelogs.html)
  'gatelogs'
  # [tcpdump](../producers/tcpdump/lib/omg-tcpdump.html)
  'tcpdump'
  # [memoryusage](../producers/memoryusage/lib/omg-memoryusage.html)
  'memoryusage'
  # [isdnconnectionsstate](../producers/isdnconnectionsstate/lib/omg-isdnconnectionsstate.html)
  'isdnconnectionsstate'
  # [dnsconfig](../producers/dnsconfig/lib/omg-dnsconfig.html)
  'dnsconfig'
  # [iproutingtable](../producers/iproutingtable/lib/omg-iproutingtable.html)
  'iproutingtable'
  # [scansipdestinations](../producers/scansipdestinations/lib/omg-scansipdestinations.html)
  'scansipdestinations'
  # [testcall](../producers/testcall/lib/omg-testcall.html)
  'testcall'
  # [testfax](../producers/testfax/lib/omg-testfax.html)
  'testfax'
  # [uptime](../producers/uptime/lib/omg-uptime.html)
  'uptime'
  # [systeminfo](../producers/systeminfo/lib/omg-systeminfo.html)
  'systeminfo'
  # [licenses](../producers/licenses/lib/omg-licenses.html)
  'licenses'
  # [callinfo](../producers/callinfo/lib/omg-callinfo.html)
  'callinfo'
  # [hostinspection](../producers/hostinspection/lib/omg-hostinspection.html)
  'hostinspection'
  # [configurator](../producers/configurator/lib/omg-configurator.html)
  'configurator'
  # [messagingservers](../producers/messagingservers/lib/omg-messagingservers.html)
  'messagingservers'
  # [numberlist](../producers/numberlist/)
  'numberlist'
  # [plain](../producers/plain/lib/omg-plain.html)
  'plain'
]

# Producers that can/must be instantiated.
producerClasses = [
  # [channels](../producers/channels/lib/omg-channels.html)
  "channels"
]

# #### Handle subscriptions from producer to producer
# # possible memory leak if not handled carefully!!!
handleInternalSubscription = (producer, callback) ->
  if omg[producer]
    omg[producer].on "publish", (messageArray) ->
      # producers publish a zmq conform array
      error = messageArray[1]
      status = messageArray[2]
      data = messageArray[3]
      try
        data = JSON.parse(data)
      catch
        data = ""
      callback error, status, data

    amountOfListeners = omg[producer].listeners("publish").length
    debug "listeners on #{producer} publish: #{amountOfListeners}"
    if amountOfListeners > 5
      console.log chalk.red.bold "#{producer} has #{amountOfListeners} listeners. Possible memory leak!!"

  # if the producer to subscribe to is not defined, retry subscribing in one second
  else
    debug "#{producer} is not defined, yet. Retrying subscription in a second"
    setTimeout ->
      handleInternalSubscription producer, callback
    , 1000

# this function adds event handlers for requests and publish events
attachEventHandlers = (producerName, producer) ->
  producer.on "request", handleRequest
  producer.on "subscribes", (to, callback) ->
    debug producerName, "subscribes to", to
    handleInternalSubscription to, callback

  producer.on "publish", (args) ->
    publishSocket.send args

# require all "normal" producers
_.each selfRunningProducers, (producerName) ->
  omg[producerName] = require "../producers/#{producerName}"
  attachEventHandlers producerName, omg[producerName]

# require and instantiate all producers that export a class
_.each producerClasses, (producerName) ->
  ProdClass = require "../producers/#{producerName}"
  omg[producerName] = new ProdClass()
  attachEventHandlers producerName, omg[producerName]
